@@ -33,7 +33,7 @@
  * Changelog:
  *   0.0.6 Added support for different sample formats.
  *
- * speaker-test.c,v 1.13 2005/11/08 16:43:36 tiwai Exp
+ * speaker-test.c,v 1.26 2006/04/06 17:24:12 tiwai Exp
  */
 
 #include <stdio.h>
@@ -43,7 +43,11 @@
 #include <errno.h>
 #include <getopt.h>
 #include <inttypes.h>
+#include <ctype.h>
+#ifdef ENABLE_NLS
 #include <locale.h>
+#endif
+#include <byteswap.h>
 
 #define ALSA_PCM_NEW_HW_PARAMS_API
 #define ALSA_PCM_NEW_SW_PARAMS_API
@@ -53,22 +57,33 @@
 #include "pink.h"
 #include "aconfig.h"
 #include "gettext.h"
+#include "version.h"
 
-static char              *device      = "plughw:0,0";       /* playback device */
+enum {
+  TEST_PINK_NOISE = 1,
+  TEST_SINE,
+  TEST_WAV
+};
+
+#define MAX_CHANNELS	16
+
+static char              *device      = "default";       /* playback device */
 static snd_pcm_format_t   format      = SND_PCM_FORMAT_S16; /* sample format */
 static unsigned int       rate        = 48000;	            /* stream rate */
 static unsigned int       channels    = 1;	            /* count of channels */
 static unsigned int       speaker     = 0;	            /* count of channels */
-static unsigned int       buffer_time = 500000;	            /* ring buffer length in us */
-static unsigned int       period_time = 100000;	            /* period time in us */
-#define PERIODS 4
+static unsigned int       buffer_time = 0;	            /* ring buffer length in us */
+static unsigned int       period_time = 0;	            /* period time in us */
+static unsigned int       nperiods    = 4;                  /* number of periods */
 static double             freq        = 440;                /* sinusoidal wave frequency in Hz */
-static int                test_type   = 1;                  /* Test type. 1 = noise, 2 = sine wave */
+static int                test_type   = TEST_PINK_NOISE;    /* Test type. 1 = noise, 2 = sine wave */
 static pink_noise_t pink;
-static snd_output_t      *output      = NULL;
 static snd_pcm_uframes_t  buffer_size;
 static snd_pcm_uframes_t  period_size;
-static const char        *channel_name[] = {
+static const char *given_test_wav_file = NULL;
+static char *wav_file_dir = DATADIR;
+
+static const char        *channel_name[MAX_CHANNELS] = {
   N_("Front Left"),
   N_("Front Right"),
   N_("Rear Left"),
@@ -77,15 +92,14 @@ static const char        *channel_name[] = {
   N_("LFE"),
   N_("Side Left"),
   N_("Side Right"),
-  N_("8"),
-  N_("9"),
-  N_("10"),
-  N_("11"),
-  N_("12"),
-  N_("13"),
-  N_("14"),
-  N_("15"),
-  N_("16")
+  N_("Channel 9"),
+  N_("Channel 10"),
+  N_("Channel 11"),
+  N_("Channel 12"),
+  N_("Channel 13"),
+  N_("Channel 14"),
+  N_("Channel 15"),
+  N_("Channel 16")
 };
 
 static const int	channels4[] = {
@@ -253,14 +267,11 @@ static void generate_pink_noise( uint8_t *frames, int channel, int count) {
 
 static int set_hwparams(snd_pcm_t *handle, snd_pcm_hw_params_t *params, snd_pcm_access_t access) {
   unsigned int rrate;
-  int          err, dir;
+  int          err;
   snd_pcm_uframes_t     period_size_min;
   snd_pcm_uframes_t     period_size_max;
   snd_pcm_uframes_t     buffer_size_min;
   snd_pcm_uframes_t     buffer_size_max;
-  snd_pcm_uframes_t     buffer_time_to_size;
-
-
 
   /* choose all parameters */
   err = snd_pcm_hw_params_any(handle, params);
@@ -305,47 +316,51 @@ static int set_hwparams(snd_pcm_t *handle, snd_pcm_hw_params_t *params, snd_pcm_
 
   printf(_("Rate set to %iHz (requested %iHz)\n"), rrate, rate);
   /* set the buffer time */
-  buffer_time_to_size = ( (snd_pcm_uframes_t)buffer_time * rate) / 1000000;
   err = snd_pcm_hw_params_get_buffer_size_min(params, &buffer_size_min);
   err = snd_pcm_hw_params_get_buffer_size_max(params, &buffer_size_max);
-  dir=0;
-  err = snd_pcm_hw_params_get_period_size_min(params, &period_size_min,&dir);
-  dir=0;
-  err = snd_pcm_hw_params_get_period_size_max(params, &period_size_max,&dir);
+  err = snd_pcm_hw_params_get_period_size_min(params, &period_size_min, NULL);
+  err = snd_pcm_hw_params_get_period_size_max(params, &period_size_max, NULL);
   printf(_("Buffer size range from %lu to %lu\n"),buffer_size_min, buffer_size_max);
   printf(_("Period size range from %lu to %lu\n"),period_size_min, period_size_max);
-  printf(_("Periods = %d\n"), PERIODS);
-  printf(_("Buffer time size %lu\n"),buffer_time_to_size);
-
-  buffer_size = buffer_time_to_size;
-  //buffer_size=8096;
-  //buffer_size=15052;
-  buffer_size=buffer_size_max;
-  if (buffer_size_max < buffer_size) buffer_size = buffer_size_max;
-  if (buffer_size_min > buffer_size) buffer_size = buffer_size_min;
-  //buffer_size=0x800;
-  period_size = buffer_size/PERIODS;
-  buffer_size = period_size*PERIODS;
-  //period_size = 510;
-  printf(_("To choose buffer_size = %lu\n"),buffer_size);
-  printf(_("To choose period_size = %lu\n"),period_size);
-  dir=0;
-  err = snd_pcm_hw_params_set_period_size_near(handle, params, &period_size, &dir);
-  if (err < 0) {
-    printf(_("Unable to set period size %lu for playback: %s\n"), period_size, snd_strerror(err));
-    return err;
+  if (period_time > 0) {
+    printf(_("Requested period time %u us\n"), period_time);
+    err = snd_pcm_hw_params_set_period_time_near(handle, params, &period_time, NULL);
+    if (err < 0) {
+      printf(_("Unable to set period time %u us for playback: %s\n"),
+	     period_time, snd_strerror(err));
+      return err;
+    }
   }
-  dir=0;
-  err = snd_pcm_hw_params_get_period_size(params, &period_size, &dir);
-  if (err < 0)  printf(_("Unable to get period size for playback: %s\n"), snd_strerror(err));
-                                                                                                                             
-  dir=0;
-  err = snd_pcm_hw_params_set_buffer_size_near(handle, params, &buffer_size);
-  if (err < 0) {
-    printf(_("Unable to set buffer size %lu for playback: %s\n"), buffer_size, snd_strerror(err));
-    return err;
+  if (buffer_time > 0) {
+    printf(_("Requested buffer time %u us\n"), buffer_time);
+    err = snd_pcm_hw_params_set_buffer_time_near(handle, params, &buffer_time, NULL);
+    if (err < 0) {
+      printf(_("Unable to set buffer time %u us for playback: %s\n"),
+	     buffer_time, snd_strerror(err));
+      return err;
+    }
   }
-  err = snd_pcm_hw_params_get_buffer_size(params, &buffer_size);
+  if (! buffer_time && ! period_time) {
+    buffer_size = buffer_size_max;
+    printf(_("Using max buffer size %lu\n"), buffer_size);
+    err = snd_pcm_hw_params_set_buffer_size_near(handle, params, &buffer_size);
+    if (err < 0) {
+      printf(_("Unable to set buffer size %lu for playback: %s\n"),
+	     buffer_size, snd_strerror(err));
+      return err;
+    }
+  }
+  if (! buffer_time || ! period_time) {
+    printf(_("Periods = %u\n"), nperiods);
+    err = snd_pcm_hw_params_set_periods_near(handle, params, &nperiods, NULL);
+    if (err < 0) {
+      printf(_("Unable to set nperiods %u for playback: %s\n"),
+	     nperiods, snd_strerror(err));
+      return err;
+    }
+  }
+  snd_pcm_hw_params_get_buffer_size(params, &buffer_size);
+  snd_pcm_hw_params_get_period_size(params, &period_size, NULL);
   printf(_("was set period_size = %lu\n"),period_size);
   printf(_("was set buffer_size = %lu\n"),buffer_size);
   if (2*period_size > buffer_size) {
@@ -353,7 +368,6 @@ static int set_hwparams(snd_pcm_t *handle, snd_pcm_hw_params_t *params, snd_pcm_
     return err;
   }
 
-
   /* write the parameters to device */
   err = snd_pcm_hw_params(handle, params);
   if (err < 0) {
@@ -433,46 +447,254 @@ static int xrun_recovery(snd_pcm_t *handle, int err) {
   return err;
 }
 
+/*
+ * Handle WAV files
+ */
+
+static const char *wav_file[MAX_CHANNELS];
+static int wav_file_size[MAX_CHANNELS];
+
+struct wave_header {
+  struct {
+    uint32_t magic;
+    uint32_t length;
+    uint32_t type;
+  } hdr;
+  struct {
+    uint32_t type;
+    uint32_t length;
+  } chunk1;
+  struct {
+    uint16_t format;
+    uint16_t channels;
+    uint32_t rate;
+    uint32_t bytes_per_sec;
+    uint16_t sample_size;
+    uint16_t sample_bits;
+  } body;
+  struct {
+    uint32_t type;
+    uint32_t length;
+  } chunk;
+};
+
+#if __BYTE_ORDER == __LITTLE_ENDIAN
+#define COMPOSE_ID(a,b,c,d)	((a) | ((b)<<8) | ((c)<<16) | ((d)<<24))
+#define LE_SHORT(v)		(v)
+#define LE_INT(v)		(v)
+#else
+#define COMPOSE_ID(a,b,c,d)	((d) | ((c)<<8) | ((b)<<16) | ((a)<<24))
+#define LE_SHORT(v)		bswap_16(v)
+#define LE_INT(v)		bswap_32(v)
+#endif
+
+#define WAV_RIFF		COMPOSE_ID('R','I','F','F')
+#define WAV_WAVE		COMPOSE_ID('W','A','V','E')
+#define WAV_FMT			COMPOSE_ID('f','m','t',' ')
+#define WAV_DATA		COMPOSE_ID('d','a','t','a')
+#define WAV_PCM_CODE		1
+
+static const char *search_for_file(const char *name)
+{
+  char *file;
+  if (*name == '/')
+    return strdup(name);
+  file = malloc(strlen(wav_file_dir) + strlen(name) + 2);
+  if (file)
+    sprintf(file, "%s/%s", wav_file_dir, name);
+  return file;
+}
+
+static int check_wav_file(int channel, const char *name)
+{
+  struct wave_header header;
+  int fd;
+
+  wav_file[channel] = search_for_file(name);
+  if (! wav_file[channel]) {
+    fprintf(stderr, _("No enough memory\n"));
+    return -ENOMEM;
+  }
+
+  if ((fd = open(wav_file[channel], O_RDONLY)) < 0) {
+    fprintf(stderr, _("Cannot open WAV file %s\n"), wav_file[channel]);
+    return -EINVAL;
+  }
+  if (read(fd, &header, sizeof(header)) < (int)sizeof(header)) {
+    fprintf(stderr, _("Invalid WAV file %s\n"), wav_file[channel]);
+    goto error;
+  }
+  
+  if (header.hdr.magic != WAV_RIFF || header.hdr.type != WAV_WAVE) {
+    fprintf(stderr, _("Not a WAV file: %s\n"), wav_file[channel]);
+    goto error;
+  }
+  if (header.body.format != LE_SHORT(WAV_PCM_CODE)) {
+    fprintf(stderr, _("Unsupported WAV format %d for %s\n"),
+	    LE_SHORT(header.body.format), wav_file[channel]);
+    goto error;
+  }
+  if (header.body.channels != LE_SHORT(1)) {
+    fprintf(stderr, _("%s is not a mono stream (%d channels)\n"),
+	    wav_file[channel], LE_SHORT(header.body.channels)); 
+    goto error;
+  }
+  if (header.body.rate != LE_INT(rate)) {
+    fprintf(stderr, _("Sample rate doesn't match (%d) for %s\n"),
+	    LE_INT(header.body.rate), wav_file[channel]);
+    goto error;
+  }
+  if (header.body.sample_bits != LE_SHORT(16)) {
+    fprintf(stderr, _("Unsupported sample format bits %d for %s\n"),
+	    LE_SHORT(header.body.sample_bits), wav_file[channel]);
+    goto error;
+  }
+  if (header.chunk.type != WAV_DATA) {
+    fprintf(stderr, _("Invalid WAV file %s\n"), wav_file[channel]);
+    goto error;
+  }
+  wav_file_size[channel] = LE_INT(header.chunk.length);
+  close(fd);
+  return 0;
+
+ error:
+  close(fd);
+  return -EINVAL;
+}
+
+static int setup_wav_file(int chn)
+{
+  static const char *wavs[MAX_CHANNELS] = {
+    "Front_Left.wav",
+    "Front_Right.wav",
+    "Rear_Left.wav",
+    "Rear_Right.wav",
+    "Front_Center.wav",
+    "Rear_Center.wav", /* FIXME: should be "Bass" or so */
+    "Side_Left.wav",
+    "Side_Right.wav",
+    "Channel_9.wav",
+    "Channel_10.wav",
+    "Channel_11.wav",
+    "Channel_12.wav",
+    "Channel_13.wav",
+    "Channel_14.wav",
+    "Channel_15.wav",
+    "Channel_16.wav"
+  };
+
+  if (given_test_wav_file)
+    return check_wav_file(chn, given_test_wav_file);
+  else
+    return check_wav_file(chn, wavs[chn]);
+}
+
+static int read_wav(uint16_t *buf, int channel, int offset, int bufsize)
+{
+  static FILE *wavfp = NULL;
+  int size;
+
+  if (! wav_file[channel]) {
+    fprintf(stderr, _("Undefined channel %d\n"), channel);
+    return -EINVAL;
+  }
+
+  if (offset >= wav_file_size[channel])
+   return 0; /* finished */
+
+  if (! offset) {
+    if (wavfp)
+      fclose(wavfp);
+    wavfp = fopen(wav_file[channel], "r");
+    if (! wavfp)
+      return -errno;
+    if (fseek(wavfp, sizeof(struct wave_header), SEEK_SET) < 0)
+      return -errno;
+  }
+  if (offset + bufsize > wav_file_size[channel])
+    bufsize = wav_file_size[channel] - offset;
+  bufsize /= channels;
+  for (size = 0; size < bufsize; size += 2) {
+    int chn;
+    for (chn = 0; chn < channels; chn++) {
+      if (chn == channel) {
+	if (fread(buf, 2, 1, wavfp) != 1)
+	  return size;
+      }
+      else
+	*buf = 0;
+      buf++;
+    }
+  }
+  return size;
+}
+
+
 /*
  *   Transfer method - write only
  */
 
-static int write_loop(snd_pcm_t *handle, int channel, int periods, uint8_t *frames) {
+static int write_buffer(snd_pcm_t *handle, uint8_t *ptr, int cptr)
+{
+  int err;
+
+  while (cptr > 0) {
+
+    err = snd_pcm_writei(handle, ptr, cptr);
+
+    if (err == -EAGAIN)
+      continue;
+
+    if (err < 0) {
+      printf(_("Write error: %d,%s\n"), err, snd_strerror(err));
+      if (xrun_recovery(handle, err) < 0) {
+	printf(_("xrun_recovery failed: %d,%s\n"), err, snd_strerror(err));
+	return -1;
+      }
+      break;	/* skip one period */
+    }
+
+    ptr += snd_pcm_frames_to_bytes(handle, err);
+    cptr -= err;
+  }
+  return 0;
+}
+
+static int write_loop(snd_pcm_t *handle, int channel, int periods, uint8_t *frames)
+{
   double phase = 0;
-  uint8_t *ptr;
-  int    err, cptr, n;
-  int    bytes_per_frame=snd_pcm_frames_to_bytes(handle, 1);
+  int    err, n;
+
+  if (test_type == TEST_WAV) {
+    int bufsize = snd_pcm_frames_to_bytes(handle, period_size);
+    n = 0;
+    while ((err = read_wav((uint16_t *)frames, channel, n, bufsize)) > 0) {
+      n += err;
+      if ((err = write_buffer(handle, frames,
+			      snd_pcm_bytes_to_frames(handle, err * channels))) < 0)
+	break;
+    }
+    if (buffer_size > n) {
+      snd_pcm_drain(handle);
+      snd_pcm_prepare(handle);
+    }
+    return err;
+  }
+    
 
   for(n = 0; n < periods; n++) {
-    if(test_type==1)
+    if (test_type == TEST_PINK_NOISE)
       generate_pink_noise(frames, channel, period_size);
     else
       generate_sine(frames, channel, period_size, &phase);
-      
-    ptr = frames;
-    cptr = period_size;
 
-    while (cptr > 0) {
-
-      err = snd_pcm_writei(handle, ptr, cptr);
-
-      if (err == -EAGAIN)
-        continue;
-
-      if (err < 0) {
-        printf(_("Write error: %d,%s\n"), err, snd_strerror(err));
-        if (xrun_recovery(handle, err) < 0) {
-          printf(_("xrun_recovery failed: %d,%s\n"), err, snd_strerror(err));
-	  return -1;
-        }
-        break;	/* skip one period */
-      }
-
-      ptr += (err * bytes_per_frame);
-      cptr -= err;
-    }
+    if ((err = write_buffer(handle, frames, period_size)) < 0)
+      return err;
+  }
+  if (buffer_size > n * period_size) {
+    snd_pcm_drain(handle);
+    snd_pcm_prepare(handle);
   }
-
   return 0;
 }
 
@@ -490,8 +712,12 @@ static void help(void)
 	   "-F,--format	sample format\n"
 	   "-b,--buffer	ring buffer size in us\n"
 	   "-p,--period	period size in us\n"
-	   "-t,--test	1=use pink noise, 2=use sine wave\n"
-	   "-s,--speaker	single speaker test. Values 1=Left or 2=right\n"
+	   "-P,--nperiods	number of periods\n"
+	   "-t,--test	pink=use pink noise, sine=use sine wave, wav=WAV file\n"
+	   "-l,--nloops	specify number of loops to test, 0 = infinite\n"
+	   "-s,--speaker	single speaker test. Values 1=Left, 2=right, etc\n"
+	   "-w,--wavfile	Use the given WAV file as a test sound\n"
+	   "-W,--wavdir	Specify the directory containing WAV files\n"
 	   "\n"));
 #if 1
   printf(_("Recognized sample formats are:"));
@@ -514,6 +740,7 @@ int main(int argc, char *argv[]) {
   uint8_t              *frames;
   int                   chn;
   double		time1,time2,time3;
+  unsigned int		n, nloops;
   struct   timeval	tv1,tv2;
 
   struct option         long_option[] = {
@@ -525,8 +752,12 @@ int main(int argc, char *argv[]) {
     {"format",    1, NULL, 'F'},
     {"buffer",    1, NULL, 'b'},
     {"period",    1, NULL, 'p'},
+    {"nperiods",  1, NULL, 'P'},
     {"test",      1, NULL, 't'},
+    {"nloops",    1, NULL, 'l'},
     {"speaker",   1, NULL, 's'},
+    {"wavfile",   1, NULL, 'w'},
+    {"wavdir",    1, NULL, 'W'},
     {NULL,        0, NULL, 0  },
   };
 
@@ -538,13 +769,14 @@ int main(int argc, char *argv[]) {
   snd_pcm_hw_params_alloca(&hwparams);
   snd_pcm_sw_params_alloca(&swparams);
  
+  nloops = 0;
   morehelp = 0;
 
-  printf("\nspeaker-test %s\n\n", SPEAKER_TEST_VERSION);
+  printf("\nspeaker-test %s\n\n", SND_UTIL_VERSION_STR);
   while (1) {
     int c;
     
-    if ((c = getopt_long(argc, argv, "hD:r:c:f:F:b:p:t:s:", long_option, NULL)) < 0)
+    if ((c = getopt_long(argc, argv, "hD:r:c:f:F:b:p:P:t:l:s:w:W:", long_option, NULL)) < 0)
       break;
     
     switch (c) {
@@ -574,18 +806,39 @@ int main(int argc, char *argv[]) {
       break;
     case 'b':
       buffer_time = atoi(optarg);
-      buffer_time = buffer_time < 1000 ? 1000 : buffer_time;
       buffer_time = buffer_time > 1000000 ? 1000000 : buffer_time;
       break;
     case 'p':
       period_time = atoi(optarg);
-      period_time = period_time < 1000 ? 1000 : period_time;
       period_time = period_time > 1000000 ? 1000000 : period_time;
       break;
+    case 'P':
+      nperiods = atoi(optarg);
+      if (nperiods < 2 || nperiods > 1024) {
+	fprintf(stderr, _("Invalid number of periods %d\n"), nperiods);
+	exit(1);
+      }
+      break;
     case 't':
-      test_type = atoi(optarg);
-      test_type = test_type < 1 ? 1 : test_type;
-      test_type = test_type > 2 ? 2 : test_type;
+      if (*optarg == 'p')
+	test_type = TEST_PINK_NOISE;
+      else if (*optarg == 's')
+	test_type = TEST_SINE;
+      else if (*optarg == 'w')
+	test_type = TEST_WAV;
+      else if (isdigit(*optarg)) {
+	test_type = atoi(optarg);
+	if (test_type < TEST_PINK_NOISE || test_type > TEST_WAV) {
+	  fprintf(stderr, _("Invalid test type %s\n"), optarg);
+	  exit(1);
+	}
+      } else {
+	fprintf(stderr, _("Invalid test type %s\n"), optarg);
+	exit(1);
+      }
+      break;
+    case 'l':
+      nloops = atoi(optarg);
       break;
     case 's':
       speaker = atoi(optarg);
@@ -596,6 +849,12 @@ int main(int argc, char *argv[]) {
         exit(EXIT_FAILURE);
       }  
       break;
+    case 'w':
+      given_test_wav_file = optarg;
+      break;
+    case 'W':
+      wav_file_dir = optarg;
+      break;
     default:
       printf(_("Unknown option '%c'\n"), c);
       exit(EXIT_FAILURE);
@@ -608,20 +867,24 @@ int main(int argc, char *argv[]) {
     exit(EXIT_SUCCESS);
   }
 
-  err = snd_output_stdio_attach(&output, stdout, 0);
-  if (err < 0) {
-    printf(_("Output failed: %s\n"), snd_strerror(err));
-    exit(EXIT_FAILURE);
-  }
+  if (test_type == TEST_WAV)
+    format = SND_PCM_FORMAT_S16_LE; /* fixed format */
 
   printf(_("Playback device is %s\n"), device);
   printf(_("Stream parameters are %iHz, %s, %i channels\n"), rate, snd_pcm_format_name(format), channels);
-  if(test_type==1)
+  switch (test_type) {
+  case TEST_PINK_NOISE:
     printf(_("Using 16 octaves of pink noise\n"));
-  else
+    break;
+  case TEST_SINE:
     printf(_("Sine wave rate is %.4fHz\n"), freq);
+    break;
+  case TEST_WAV:
+    printf(_("WAV file(s)\n"));
+    break;
+
+  }
 
-loop:
   while ((err = snd_pcm_open(&handle, device, SND_PCM_STREAM_PLAYBACK, 0)) < 0) {
     printf(_("Playback open error: %d,%s\n"), err,snd_strerror(err));
     sleep(1);
@@ -630,26 +893,32 @@ loop:
   if ((err = set_hwparams(handle, hwparams, SND_PCM_ACCESS_RW_INTERLEAVED)) < 0) {
     printf(_("Setting of hwparams failed: %s\n"), snd_strerror(err));
     snd_pcm_close(handle);
-    goto loop;
     exit(EXIT_FAILURE);
   }
-  //getchar();
   if ((err = set_swparams(handle, swparams)) < 0) {
     printf(_("Setting of swparams failed: %s\n"), snd_strerror(err));
     snd_pcm_close(handle);
-    goto loop;
     exit(EXIT_FAILURE);
   }
 
-  frames = malloc((period_size * channels * snd_pcm_format_width(format)) / 8);
-  initialize_pink_noise( &pink, 16);
+  frames = malloc(snd_pcm_frames_to_bytes(handle, period_size));
+  if (test_type == TEST_PINK_NOISE)
+    initialize_pink_noise(&pink, 16);
   
   if (frames == NULL) {
     printf(_("No enough memory\n"));
     exit(EXIT_FAILURE);
   }
   if (speaker==0) {
-    while (1) {
+
+    if (test_type == TEST_WAV) {
+      for (chn = 0; chn < channels; chn++) {
+	if (setup_wav_file(chn) < 0)
+	  exit(EXIT_FAILURE);
+      }
+    }
+
+    for (n = 0; ! nloops || n < nloops; n++) {
 
       gettimeofday(&tv1, NULL);
       for(chn = 0; chn < channels; chn++) {
@@ -666,18 +935,12 @@ loop:
         printf(" %d - %s\n", channel, gettext(channel_name[channel]));
 
         err = write_loop(handle, channel, ((rate*3)/period_size), frames);
-        //err = write_loop(handle, 255, ((rate*3)/period_size), frames);
 
         if (err < 0) {
           printf(_("Transfer failed: %s\n"), snd_strerror(err));
           free(frames);
           snd_pcm_close(handle);
-	  printf(_("Pausing\n"));
-	  goto loop ;
-	  //pause();
-	  //printf("Done Pausing\n");
           exit(EXIT_SUCCESS);
-	  goto loop ;
         }
       }
       gettimeofday(&tv2, NULL);
@@ -687,6 +950,11 @@ loop:
       printf(_("Time per period = %lf\n"), time3 );
     }
   } else {
+    if (test_type == TEST_WAV) {
+      if (setup_wav_file(speaker - 1) < 0)
+	exit(EXIT_FAILURE);
+    }
+
     printf("  - %s\n", gettext(channel_name[speaker-1]));
     err = write_loop(handle, speaker-1, ((rate*5)/period_size), frames);
 
